from copy import copy
from pathlib import Path

import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS

from sunpy import log as logger
from sunpy.coordinates.ephemeris import get_body_heliographic_stonyhurst
from sunpy.coordinates.frames import HeliographicStonyhurst, Helioprojective
from sunpy.coordinates.screens import SphericalScreen
from sunpy.coordinates.wcs_utils import _set_wcs_aux_obs_coord

from irispy.spectrograph import Collection, SGMeta, SpectrogramCube, SpectrogramCubeSequence
from irispy.utils import calculate_uncertainty
from irispy.utils.constants import DN_UNIT, READOUT_NOISE, SLIT_WIDTH

__all__ = ["read_spectrograph_lvl2"]


def _pc_matrix(lam, angle_1, angle_2):
    return angle_1, -1 * lam * angle_2, 1 / lam * angle_2, angle_1


def read_spectrograph_lvl2(
    filenames,
    *,
    spectral_windows=None,
    uncertainty=False,
    memmap=False,
    revert_v34=False,
):
    """
    Reads either a SINGLE IRIS level 2 spectrograph FITs or a list of them.

    Does not handle tar files.

    Parameters
    ----------
    filenames: `list` of `str` or `str`
        Filename of filenames to be read. They must all be associated with the same
        OBS number.
    spectral_windows: iterable of `str` or `str`
        Spectral windows to extract from files. Default=None, implies, extract all
        spectral windows.
    uncertainty : `bool`, optional
        If `True` (not the default), will compute the uncertainty for the data (slower and
        uses more memory). If `memmap=True`, the uncertainty is never computed.
    memmap : `bool`, optional
        If `True` (not the default), will not load arrays into memory, and will only read from
        the file into memory when needed. This option is faster and uses a
        lot less memory. However, because FITS scaling is not done on-the-fly,
        the data units will be unscaled, not the usual data numbers (DN).
    revert_v34 : `bool`, optional.
        Will undo the data and WCS flipping made to V34 observations.
        Defaults to `False`.

    Returns
    -------
    `ndcube.NDCollection`
    """
    if isinstance(filenames, (str, Path)):
        filenames = [filenames]
    filenames = [str(f) for f in filenames]
    # Collecting the window observations
    with fits.open(filenames[0], memmap=memmap, do_not_scale_image_data=memmap) as hdulist:
        # After a discussion with the IRIS team, it was decided that instead of the
        # OBSID, we will use STEPS_AV less than -0.01 to identify V34 observations.
        v34 = hdulist[0].header["STEPS_AV"] < -0.01
        hdulist.verify("silentfix")
        windows_in_obs = np.array(
            [hdulist[0].header[f"TDESC{i}"] for i in range(1, hdulist[0].header["NWIN"] + 1)],
        )
        # If spectral_window is not set then get every window.
        # Else take the appropriate windows
        if not spectral_windows:
            spectral_windows_req = windows_in_obs
            window_fits_indices = range(1, len(hdulist) - 2)
        else:
            spectral_windows_req = [spectral_windows] if isinstance(spectral_windows, str) else spectral_windows
            spectral_windows_req = np.asarray(spectral_windows_req, dtype="U")
            window_is_in_obs = np.asarray([window in windows_in_obs for window in spectral_windows_req])
            if not all(window_is_in_obs):
                missing_windows = window_is_in_obs is False
                msg = f"Spectral windows {spectral_windows[missing_windows]} not in file {filenames[0]}"
                raise ValueError(msg)
            window_fits_indices = np.nonzero(np.isin(windows_in_obs, spectral_windows))[0] + 1
        data_dict = {window_name: [] for window_name in spectral_windows_req}
        # No observer information in the header, so we just assume its at Earth.
        base_time = Time(hdulist[0].header["STARTOBS"])
        location = get_body_heliographic_stonyhurst("Earth", (base_time).isot)
        observer = Helioprojective(
            hdulist[0].header["XCEN"] * u.arcsec,
            hdulist[0].header["YCEN"] * u.arcsec,
            observer=location,
            obstime=base_time,
        )
        with SphericalScreen(observer.observer):
            observer = observer.transform_to(HeliographicStonyhurst(obstime=base_time))
    for filename in filenames:
        with fits.open(filename, memmap=memmap, do_not_scale_image_data=memmap) as hdulist:
            hdulist.verify("silentfix")
            # Extract axis-aligned metadata.
            times = Time(hdulist[0].header["STARTOBS"]) + TimeDelta(
                hdulist[-2].data[:, hdulist[-2].header["TIME"]],
                format="sec",
            )
            fov_center = SkyCoord(
                Tx=hdulist[-2].data[:, hdulist[-2].header["XCENIX"]],
                Ty=hdulist[-2].data[:, hdulist[-2].header["YCENIX"]],
                unit=u.arcsec,
                frame=Helioprojective,
            )
            obs_vrix = hdulist[-2].data[:, hdulist[-2].header["OBS_VRIX"]] * u.m / u.s
            ophaseix = hdulist[-2].data[:, hdulist[-2].header["OPHASEIX"]]
            exposure_times_fuv = hdulist[-2].data[:, hdulist[-2].header["EXPTIMEF"]] * u.s
            exposure_times_nuv = hdulist[-2].data[:, hdulist[-2].header["EXPTIMEN"]] * u.s
            for i, window_name in enumerate(spectral_windows_req):
                meta = SGMeta(
                    hdulist[0].header,
                    window_name,
                    data_shape=hdulist[window_fits_indices[i]].data.shape,
                )
                exposure_times = exposure_times_nuv
                dn_unit = DN_UNIT["NUV"]
                readout_noise = READOUT_NOISE["NUV"]
                if "FUV" in meta.detector:
                    exposure_times = exposure_times_fuv
                    dn_unit = DN_UNIT["FUV"]
                    readout_noise = READOUT_NOISE["FUV"]
                meta.add("exposure time", exposure_times, None, 0)
                meta.add("exposure FOV center", fov_center, None, 0)
                meta.add("observer radial velocity", obs_vrix, None, 0)
                meta.add("orbital phase", ophaseix, None, 0)
                # Sit-and-stare have a CDELT of 0 which causes issues in WCS.
                # In this case, set CDELT to a small number.
                header = copy(hdulist[window_fits_indices[i]].header)
                # Account for a slit offset (POFFYNUV (45) or POFFYFUV (34))
                idx = 34 if meta.spectral_band == "FUV" else 45
                header["CRVAL3"] -= hdulist[-2].data[:, idx].mean() * (SLIT_WIDTH.value / 2)
                if header["CDELT3"] == 0:
                    header["CDELT3"] = 1e-10
                    ang1, ang2, ang3, ang4 = _pc_matrix(
                        header["CDELT3"] / header["CDELT2"],
                        hdulist[-2].data[:, 20].mean(),
                        hdulist[-2].data[:, 22].mean(),
                    )
                    header["PC2_2"] = ang1
                    header["PC2_3"] = ang2
                    header["PC3_2"] = ang3
                    header["PC3_3"] = ang4
                try:
                    wcs = WCS(header)
                except Exception as e:  # NOQA: BLE001
                    msg = (
                        f"WCS failed to load while reading one step of the raster due to {e}"
                        "The loading will continue but this will be missing in the final cube. "
                        f"Spectral window: {window_name}, step {i} in file: {filename}"
                    )
                    logger.warning(msg)
                    continue
                out_uncertainty = None
                data_mask = None
                if not memmap:
                    data_mask = hdulist[window_fits_indices[i]].data == -200.0
                if uncertainty:
                    out_uncertainty = calculate_uncertainty(
                        hdulist[window_fits_indices[i]].data,
                        readout_noise,
                        dn_unit,
                    )
                if v34 and not revert_v34:
                    times = times[::-1]
                    data = np.flip(hdulist[window_fits_indices[i]].data, axis=0)
                    header["PC1_3"] = 0
                    header["PC2_3"] = -header["PC2_3"]
                    header["PC3_2"] = -header["PC3_2"]
                    header["CDELT3"] = -header["CDELT3"]
                    header["CRPIX3"] = header["NAXIS3"] - header["CRPIX3"] + 1
                    wcs = WCS(header)
                else:
                    data = hdulist[window_fits_indices[i]].data
                _set_wcs_aux_obs_coord(wcs, observer)
                cube = SpectrogramCube(
                    data,
                    wcs=wcs,
                    uncertainty=out_uncertainty,
                    unit=dn_unit,
                    meta=meta,
                    mask=data_mask,
                )
                cube.extra_coords.add("time", 0, times, physical_types="time")
                data_dict[window_name].append(cube)
    window_data_pairs = [
        (window_name, SpectrogramCubeSequence(data_dict[window_name], common_axis=0))
        for window_name in spectral_windows_req
    ]
    return Collection(window_data_pairs, aligned_axes=(0, 1, 2))



"""
==================================
Opening the IRIS Aligned AIA Cubes
==================================

In this example we will show how ``irispy-lmsal`` handles the AIA cubes provided by the IRIS team.
These cubes are aligned to the IRIS observation and are 50" larger than the IRIS FOV.

They have the same format as IRIS SJI files, so you can read them via ``irispy-lmsal``.
"""

import matplotlib.pyplot as plt
import pooch

from irispy.io import read_files

###############################################################################
# We start by downloading an AIA cube from the IRIS data archive.
#
# In this case, we will use ``pooch`` to keep this example self-contained
# but using your browser will also work.

sdo_aia_file = pooch.retrieve(
    "https://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2025/05/19/20250519_165924_3640107442/iris_l2_20250519_165924_3640107442_SDO.tar.gz",
    known_hash="b77d693fa328a96aec78b4a4aa420d5b167f8f670719fce815836627ed567f42",
)

###############################################################################
# We will now open the AIA dataset.
# It is provided as a compressed archive, with each AIA wavelength as a separate
# file.
#
# In this example, they are:
#
# - aia_l2_20250519_165924_3640107442_171.fits
# - aia_l2_20250519_165924_3640107442_94.fits
# - aia_l2_20250519_165924_3640107442_304.fits
# - aia_l2_20250519_165924_3640107442_193.fits
# - aia_l2_20250519_165924_3640107442_335.fits
# - aia_l2_20250519_165924_3640107442_211.fits
# - aia_l2_20250519_165924_3640107442_1700.fits
# - aia_l2_20250519_165924_3640107442_131.fits
# - aia_l2_20250519_165924_3640107442_1600.fits

# This will return a list of the AIA cubes.
aia_collection = read_files(sdo_aia_file, memmap=False)

###############################################################################
# Let us look at the first collection returned of the AIA cube.

print(aia_collection)

###############################################################################
# We will then select the 304 bandpass cube.

print(aia_collection["304_THIN"])

###############################################################################
# We will now plot the AIA data in the same manner as the SJI data.
#
# You can also change the axis labels and ticks if you so desire.
# `WCSAxes provides us an API we can use. <https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html>`__

fig = plt.figure()
# Note that the .get_animation() is used to animate this example
# and is not required normally.
aia_collection["304_THIN"].plot(fig=fig).get_animation()

plt.show()



"""
This module provides general utility functions called by code in spectrograph.
"""

import numpy as np

import astropy.units as u
from astropy import constants

from irispy.utils.constants import RADIANCE_UNIT
from irispy.utils.utils import get_interpolated_effective_area

__all__ = [
    "calculate_photons_per_sec_to_radiance_factor",
    "convert_between_dn_and_photons",
    "convert_or_undo_photons_per_sec_to_radiance",
    "reshape_1d_wavelength_dimensions_for_broadcast",
]


def convert_between_dn_and_photons(old_data_arrays, old_unit, new_unit):
    """
    Converts arrays from IRIS DN to photons or vice versa.

    In this function, an inverse time component due to exposure time
    correction is ignored during calculations but preserved in final unit.

    Parameters
    ----------
    old_data_arrays: iterable of `numpy.ndarray`
        Arrays of data to be converted.
    old_unit: `astropy.unit.Unit`
        Unit of data arrays.
    new_unit: `astropy.unit.Unit`
        Unit to convert data arrays to.

    Returns
    -------
    `list` of `numpy.ndarray`
        Data arrays converted to new_unit.
    `astropy.unit.Unit`
        Unit of new data arrays with any inverse time component preserved.
    """
    if old_unit in [new_unit, new_unit / u.s]:
        new_data_arrays = list(old_data_arrays)
        new_unit_time_accounted = old_unit
    else:
        # During calculations, the time component due to exposure
        # time correction, if it has been applied, is ignored.
        # Check here whether the time correction is present in the
        # original unit so that is carried through to new unit.
        if u.s not in (old_unit * u.s).decompose().bases:
            old_unit_without_time = old_unit * u.s
            new_unit_time_accounted = new_unit / u.s
        else:
            old_unit_without_time = old_unit
            new_unit_time_accounted = new_unit
        # Convert data and uncertainty to new unit.
        new_data_arrays = [(data * old_unit_without_time).to(new_unit).value for data in old_data_arrays]
    return new_data_arrays, new_unit_time_accounted


def convert_or_undo_photons_per_sec_to_radiance(
    data_quantities,
    time_obs,
    response_version,
    obs_wavelength,
    detector_type,
    spectral_dispersion_per_pixel,
    solid_angle,
    *,
    undo=False,
):
    """
    Converts data quantities from counts/s to radiance (or vice versa).

    Parameters
    ----------
    data_quantities: iterable of `astropy.units.Quantity`
        Quantities to be converted.  Must have units of counts/s or
        radiance equivalent counts, e.g. erg / cm**2 / s / sr / Angstrom.
    time_obs: an `astropy.time.Time` object, as a kwarg, valid for version > 2
        Observation times of the datapoints.
        Must be in the format of, e.g.,
        time_obs parse_time('2013-09-03', format='utime'),
        which yields 1094169600.0 seconds in value.
        The argument time_obs is ignored for versions 1 and 2.
    response_version : `int`
        Version number of effective area file to be used. Cannot be set
        simultaneously with response_file or pre_launch kwarg. Default=4.
    obs_wavelength: `astropy.units.Quantity`
        Wavelength at each element along spectral axis of data quantities.
    detector_type: `str`
        Detector type: 'FUV', 'NUV', or 'SJI'.
    spectral_dispersion_per_pixel: scalar `astropy.units.Quantity`
        spectral dispersion (wavelength width) of a pixel.
    solid_angle: scalar `astropy.units.Quantity`
        Solid angle corresponding to a pixel.
    undo: `bool`
        If False, converts counts/s to radiance.
        If True, converts radiance to counts/s.
        Default=False

    Returns
    -------
    `list` of `astropy.units.Quantity`
        Data quantities converted to radiance or counts/s
        depending on value of undo kwarg.
    """
    # Check data quantities are in the right units.
    if undo is True:
        for i, data in enumerate(data_quantities):
            if not data.unit.is_equivalent(RADIANCE_UNIT):
                msg = (
                    "Invalid unit provided.  As kwarg undo=True, "
                    f"unit must be equivalent to {RADIANCE_UNIT}.  Error found for {i}th element "
                    f"of data_quantities. Unit: {data.unit}"
                )
                raise ValueError(
                    msg,
                )
    else:
        for data in data_quantities:
            if data.unit != u.photon / u.s:
                msg = (
                    "Invalid unit provided.  As kwarg undo=False, "
                    f"unit must be equivalent to {u.photon / u.s}.  Error found for {i}th element "
                    f"of data_quantities. Unit: {data.unit}"
                )
                raise ValueError(
                    msg,
                )
    photons_per_sec_to_radiance_factor = calculate_photons_per_sec_to_radiance_factor(
        time_obs,
        response_version,
        obs_wavelength,
        detector_type,
        spectral_dispersion_per_pixel,
        solid_angle,
    )
    # Change shape of arrays so they are compatible for broadcasting
    # with data and uncertainty arrays.
    photons_per_sec_to_radiance_factor = reshape_1d_wavelength_dimensions_for_broadcast(
        photons_per_sec_to_radiance_factor,
        data_quantities[0].ndim,
    )
    return (
        [(data / photons_per_sec_to_radiance_factor).to(u.photon / u.s) for data in data_quantities]
        if undo is True
        else [(data * photons_per_sec_to_radiance_factor).to(RADIANCE_UNIT) for data in data_quantities]
    )


def calculate_photons_per_sec_to_radiance_factor(
    time_obs,
    response_version,
    wavelength,
    detector_type,
    spectral_dispersion_per_pixel,
    solid_angle,
):
    """
    Calculates multiplicative factor that converts counts/s to radiance for
    given wavelengths.

    Parameters
    ----------
    time_obs: an `astropy.time.Time` object, as a kwarg, valid for version > 2
        Observation times of the datapoints.
        Must be in the format of, e.g.,
        time_obs=parse_time('2013-09-03', format='utime'),
        which yields 1094169600.0 seconds in value.
        The argument time_obs is ignored for versions 1 and 2.
    response_version : `int`
        Version number of effective area file to be used. Cannot be set
        simultaneously with response_file or pre_launch kwarg. Default=4.
    wavelength: `astropy.units.Quantity`
        Wavelengths for which counts/s-to-radiance factor is to be calculated
    detector_type: `str`
        Detector type: 'FUV' or 'NUV'.
    spectral_dispersion_per_pixel: scalar `astropy.units.Quantity`
        spectral dispersion (wavelength width) of a pixel.
    solid_angle: scalar `astropy.units.Quantity`
        Solid angle corresponding to a pixel.

    Returns
    -------
    `astropy.units.Quantity`
        # The term "multiplicative" refers to the fact that the conversion factor calculated by the
        # `calculate_photons_per_sec_to_radiance_factor` function is used to multiply the counts per
        # second (cps) data to obtain the radiance data. In other words, the conversion factor is a
        # scaling factor that is applied to the cps data to convert it to radiance units.
        Multiplicative conversion factor from counts/s to radiance units
        for input wavelengths.
    """
    # Get effective area and interpolate to observed wavelength grid.
    eff_area_interp = get_interpolated_effective_area(
        time_obs,
        response_version,
        detector_type,
        obs_wavelength=wavelength,
    )
    # Return radiometric conversed data assuming input data is in units of photons/s.
    return (
        constants.h
        * constants.c
        / wavelength
        / u.photon
        / spectral_dispersion_per_pixel
        / eff_area_interp
        / solid_angle
    )


def reshape_1d_wavelength_dimensions_for_broadcast(wavelength, n_data_dim):
    if n_data_dim == 1:
        pass
    elif n_data_dim == 2:
        wavelength = wavelength[np.newaxis, :]
    elif n_data_dim == 3:
        wavelength = wavelength[np.newaxis, np.newaxis, :]
    else:
        msg = "IRISSpectrogram dimensions must be 2 or 3."
        raise ValueError(msg)
    return wavelength



import re
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation, patheffects

from astropy.io import fits
from astropy.time import TimeDelta
from astropy.visualization import AsinhStretch, ImageNormalize
from astropy.wcs import WCS

from sunpy.time import parse_time
from sunpy.visualization.colormaps.color_tables import iris_sji_color_table

from irispy.utils import image_clipping

__all__ = ["wobble_movie"]


def wobble_movie(
    files: list | str | Path,
    *,
    outdir: str | Path = "./",
    trim: bool = False,
    timestamp: bool = True,
    wobble_cadence: int = 180,
    ffmpeg_path: str | Path | None = None,
    **kwargs,
) -> None:
    """
    Creates a wobble movie from a list of files.

    ..note:

        This requires FFMPEG to be installed and discoverable.
        If FFMPEG is not found, you can pass it as an argument called ``ffmpeg_path``.

    Parameters
    ----------
    files : Union[list, str, Path]
        Files to create a wobble movie from.
        If a string or Path is passed, it will encapsulated in a list.
    outdir : Union[str,Path], optional
        Location to save the movie(s).
        Defaults to the current working directory.
    trim : `bool`, optional
        Movie is trimmed to include only area that has data in all frames, by default False
    timestamp : `bool`, optional
        If `True`, will add a timestamp to the wobble movie.
        Optional, defaults to `True`.
    wobble_cadence : `int`, optional
        Sets the cadence of the wobble movie in seconds.
        Optional, defaults to 180 seconds.
    ffmpeg_path : Union[str,Path], optional
        Path to FFMPEG executable, by default `None`.
        In theory you will not need to do this but matplotlib might not be able to find the ffmpeg exe.
    **kwargs : `dict`, optional
        Keyword arguments to passed to `FuncAnimation`.

    Returns
    -------
    `list`
        A list of the movies created.

    Notes
    -----
    This is designed to be used on IRIS Level 2 SJI data.

    2832 is considered the best wavelength to use for wobble movies.

    Timestamps take the main header cadence and add that to the "DATEOBS".
    They do not use the information in the AUX array.
    """
    if ffmpeg_path:
        mpl.rcParams["animation.ffmpeg_path"] = ffmpeg_path

    if isinstance(files, str | Path):
        files = [files]
    filenames = []
    for a_file in files:
        data, header = fits.getdata(a_file, header=True)
        wcs = WCS(header)
        numframes = header["NAXIS3"]
        date = header["DATE_OBS"].split(".")[0]
        # Calculate index to downsample in time to accentuate the wobble
        cadence = header["CDELT3"]
        cadence_sample = max(1, np.floor(wobble_cadence / cadence))
        if timestamp:
            timestamps = [
                parse_time(header["STARTOBS"]) + TimeDelta(cadence, format="sec") * i for i in range(numframes)
            ]
        else:
            timestamps = [parse_time(header["STARTOBS"])]
        # Trim down to only that part of the movie that contains data in all frames
        if trim:
            # TODO: improve this, it trims a bit but not fully
            dmin = np.min(data, axis=0)
            dmask = dmin > -200
            dmx = np.sum(dmask, axis=1)
            dmy = np.sum(dmask, axis=0)
            (subx,) = np.where(dmx > (np.max(dmx) * 0.8))
            (suby,) = np.where(dmy > (np.max(dmy) * 0.8))
            data = data[:, suby[0] : suby[-1], subx[0] : subx[-1]]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=wcs.dropaxis(-1))
        colormap = iris_sji_color_table(str(int(header["TWAVE1"])))
        vmin, vmax = image_clipping(data)
        image = ax.imshow(
            data[0],
            origin="lower",
            cmap=colormap,
            norm=ImageNormalize(vmin=vmin, vmax=vmax, stretch=AsinhStretch()),
        )
        ax.set_xlabel("Solar X")
        ax.set_ylabel("Solar Y")
        if timestamp:
            title = ax.text(
                0.5,
                0.95,
                str(timestamps[0]),
                color="w",
                transform=ax.transAxes,
                ha="center",
                path_effects=[patheffects.withStroke(linewidth=3, foreground="black")],
            )
        else:
            title = ax.text(0.5, 0.95, "")

        def update(i):
            image.set_array(data[i])  # NOQA: B023
            if timestamp:
                title.set_text(str(timestamps[i]))  # NOQA: B023
            return image, title  # NOQA: B023

        anim = animation.FuncAnimation(
            fig,
            func=update,
            frames=range(0, numframes, int(cadence_sample)),
            blit=True,
            repeat=False,
            **kwargs,
        )
        clean_filename = re.sub(r"[^\w\-_\. ]", "_", f"{header['TDESC1']}_{date}_wobble.mp4")
        filename = Path(outdir) / Path(clean_filename)
        writervideo = animation.FFMpegWriter(fps=12)
        anim.save(filename, writer=writervideo)
        filenames.append(filename)
    return filenames



"""
================
Spectral fitting
================

In this example, we are going to fit spectral lines from IRIS, using the raster data with a single Gaussian.
Then use the fitted values to calculate the Gaussian moments.
"""
# sphinx_gallery_thumbnail_number = 3

import warnings

import matplotlib.pyplot as plt
import numpy as np
import pooch

import astropy.units as u
from astropy import constants
from astropy.coordinates import SkyCoord, SpectralCoord
from astropy.modeling import models as m
from astropy.modeling.fitting import LevMarLSQFitter, parallel_fit_dask
from astropy.visualization import time_support
from astropy.wcs.utils import wcs_to_celestial_frame

from sunpy.coordinates.frames import Helioprojective

from irispy.io import read_files

time_support()

###############################################################################
# We will start by getting some data from the IRIS archive.
#
# In this case, we will use ``pooch`` so to keep this example self-contained
# but using your browser will also work.
#
# Using the url: http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2018/01/02/20180102_153155_3610108077/iris_l2_20180102_153155_3610108077_raster.tar.gz
# we are after the raster sequence (~300 MB).

raster_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2018/01/02/20180102_153155_3610108077/iris_l2_20180102_153155_3610108077_raster.tar.gz",
    known_hash="0ec2b7b20757c52b02e0d92c27a5852b6e28759512c3d455f8b6505d4e1f5cd6",
)

###############################################################################
# Now to open the files using ``irispy-lmsal``.

# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32 (via "b_zero" and "b_scale"),
# the data values are no longer in DN, but in scaled integer units that start at -2$^{16}$/2.
#
# We will use ``memmap=False`` because we want to fit the actual the data values.

raster = read_files(raster_filename, memmap=False)

###############################################################################
# We will just focus on the Si IV 1403 line which we can select using a key.
# Then we will just plot a spectral line selected at random in space.

# There is only one complete scan, so we index that away.
si_iv_1403 = raster["Si IV 1403"][0]

# However, before we get to that, we will shrink the data cube to make it easier to work with.
iris_observer = wcs_to_celestial_frame(si_iv_1403.wcs.celestial).observer
iris_frame = Helioprojective(observer=iris_observer)
top_left = [None, SkyCoord(-290 * u.arcsec, 260 * u.arcsec, frame=iris_frame)]
bottom_right = [None, SkyCoord(-360 * u.arcsec, 310 * u.arcsec, frame=iris_frame)]
si_iv_1403 = si_iv_1403.crop(top_left, bottom_right)

###############################################################################
# Let us just check the full field of view at the line core.

si_iv_core = 140.277 * u.nm
lower_corner = [SpectralCoord(si_iv_core), None]
upper_corner = [SpectralCoord(si_iv_core), None]
si_iv_spec_crop = si_iv_1403.crop(lower_corner, upper_corner)

fig = plt.figure()
ax = fig.add_subplot(111, projection=si_iv_spec_crop.wcs)
si_iv_spec_crop.plot(axes=ax, vmin=0, vmax=200)
plt.title("Si IV 1402.77 A")
plt.colorbar(label="Intensity [DN]", shrink=0.8)

###############################################################################
# We will want to make two rebinned cubes from the full raster,
# one summed along the wavelength dimension and one of the spectra averaged over all spatial pixels.

wl_sum = si_iv_1403.rebin((1, 1, si_iv_1403.data.shape[-1]), operation=np.sum)[0]
spatial_mean = si_iv_1403.rebin((*si_iv_1403.data.shape[:-1], 1))[0, 0, :]

################################################################################
# Now we can create a model for this spectra.

initial_model = m.Const1D(amplitude=2 * si_iv_1403.unit) + m.Gaussian1D(
    amplitude=8 * si_iv_1403.unit, mean=si_iv_core, stddev=0.005 * u.nm
)

################################################################################
# To improve our initial conditions we now fit the initial model to the spatially averaged spectra.
# To do this we use the `ndcube.NDCube.axis_world_coords` method of NDCube which returns all,
# or a subset of the world coordinates along however many array axes they are
# correlated with. So in this case we get the wavelength dimension which only
# returns a single `SpectralCoord` object corresponding to the first array dimension of the cube.

fitter = LevMarLSQFitter()
average_fit = fitter(
    initial_model,
    spatial_mean.axis_world_coords("em.wl")[0].to(u.nm),
    spatial_mean.data * spatial_mean.unit,
)

################################################################################
# Now we check, the initial model and the model fitted to the average spectra.

fig = plt.figure()
ax = spatial_mean.plot(label="Spatial average")
ax.plot(initial_model(spatial_mean.axis_world_coords("em.wl")[0].to(u.nm)), label="Initial model")
ax.plot(average_fit(spatial_mean.axis_world_coords("em.wl")[0].to(u.nm)), linestyle="--", label="Spatial average fit")
plt.legend()

################################################################################
# The function `.parallel_fit_dask` will map a model to each element of a cube along
# one (or more) "fitting axes", in this case our fitting axis is our wavelength
# axis (array axis -1). So we want to fit each slice of the data array along the 3rd axis.
#
# The key arguments to the parallel_fit_dask function are:
#
# * A data array: This can be a numpy array or a dask array, or a NDData (or subclass like NDCube)
#                 object. If it's a NDData object then the data, wcs, mask, data_unit and uncertainty
#                 are all extracted from the NDData object and used in place of their respective keyword
#                 arguments.
# * A model to fit
# * A fitter instance.
# * The fitting axis (or axes).
#
# What is returned from `.parallel_fit_dask` is a model with array parameters with
# the shape of the non-fitting axes of the data.

# We want to do some basic data sanitization.
# Remove negative values and set them to zero and remove non-finite values.
filtered_data = np.where(si_iv_1403.data < 0, 0, si_iv_1403.data)
filtered_data = np.where(np.isfinite(filtered_data), filtered_data, 0)

# We can therefore fit the cube
with warnings.catch_warnings():
    # There are several WCS warnings we just want to ignore
    warnings.simplefilter("ignore")
    iris_model_fit = parallel_fit_dask(
        data=filtered_data,
        data_unit=si_iv_1403.unit,
        fitting_axes=2,
        world=si_iv_1403.wcs,
        model=average_fit,
        fitter=LevMarLSQFitter(),
        scheduler="single-threaded",
    )

# Note that this example is done in a single thread. If you want to use multiple cores.
# You can create a dask client and pass it to the parallel_fit_dask function.
# For example:
#
# from dask.distributed import Client
#
# client = Client()
#
# Then pass this to the parallel_fit_dask function by replacing scheduler line above with:
#
# scheduler=client,
#

################################################################################
# Let us see the output!

fig, axs = plt.subplots(nrows=1, ncols=3, subplot_kw={"projection": si_iv_spec_crop}, figsize=(16, 6))
net_flux = (
    np.sqrt(2 * np.pi)
    * (iris_model_fit.amplitude_0 + iris_model_fit.amplitude_1)
    * iris_model_fit.stddev_1.quantity
    / np.mean(si_iv_1403.axis_world_coords("wl")[0][1:] - si_iv_1403.axis_world_coords("wl")[0][:-1])
)
amp_max = np.nanpercentile(np.abs(net_flux.value), 99)
amp = axs[0].imshow(net_flux.value, vmin=0, vmax=amp_max, origin="lower")
cbar = fig.colorbar(amp, ax=axs[0])
cbar.set_label(label=f"Intensity [{net_flux.unit.to_string()}]", fontsize=8)
cbar.ax.tick_params(labelsize=8)
axs[0].set_title("Gaussian Net Flux")

core_shift = ((iris_model_fit.mean_1.quantity.to(u.nm)) - si_iv_core) / si_iv_core * (constants.c.to(u.km / u.s))
shift_max = np.nanpercentile(np.abs(core_shift.value), 95)
shift = axs[1].imshow(core_shift.value, cmap="coolwarm", vmin=-shift_max, vmax=shift_max)
cbar = fig.colorbar(shift, ax=axs[1], extend="both")
cbar.set_label(label=f"Doppler shift [{core_shift.unit.to_string()}]", fontsize=8)
cbar.ax.tick_params(labelsize=8)
axs[1].set_title("Velocity from Gaussian shift")

sigma = (iris_model_fit.stddev_1.quantity.to(u.nm)) / si_iv_core * (constants.c.to(u.km / u.s))
line_max = np.nanpercentile(np.abs(sigma.value), 95)
line = axs[2].imshow(sigma.value, vmax=line_max)
cbar = fig.colorbar(line, ax=axs[2])
cbar.set_label(label=f"Line Width [{sigma.unit.to_string()}]", fontsize=8)
cbar.ax.tick_params(labelsize=8)
axs[2].set_title("Gaussian Sigma")

for ax in axs:
    ax.coords[0].set_ticklabel(exclude_overlapping=True, fontsize=8)
    ax.coords[0].set_axislabel("Helioprojective Longitude", fontsize=8)
    ax.coords[1].set_ticklabel(exclude_overlapping=True, fontsize=8)
    ax.coords[1].set_axislabel("Helioprojective Latitude", fontsize=8)
fig.tight_layout()

plt.show()



import importlib.resources
from pathlib import Path

import numpy as np
import pandas as pd

from astropy import units as u

dir_path = Path(__file__).parent

__all__ = ["ObsID"]


class ObsID(dict):
    """
    A class to convert the IRIS OBS ID to human-readable format.

    Parameters
    ----------
    obsid : `int`
        IRIS OBS ID to query. Needs to be a valid OBS ID, with 10 digits.

    Notes
    -----
    Currently, only OBS IDs with version numbers 36, 38, and 40 are
    supported. Calibration OBS IDs (version 42) are not yet supported.

    Examples
    --------
    Quickly show OBS ID parameters

    >>> from irispy import obsid
    >>> obsid.ObsID(3675508564)
    IRIS OBS ID 3675508564
    ----------------------
    Description:            Large dense 96-step raster 31.35x120 96s
    SJI filters:                                    C II   Mg II w s
    SJI field of view:                                       120x120
    Exposure time:                                             8.0 s
    Binning:                               Spatial x 1, Spectral x 1
    FUV binning:                         FUV spectrally rebinned x 4
    SJI cadence:                                 SJI cadence default
    Compression:                                Lossless compression
    Linelist:                                       Flare linelist 1

    The data can be accessed as in a dictionary:

    >>> data = obsid.ObsID(3675508564)
    >>> data["exptime"]
    <Quantity 8. s>
    >>> data["linelist"]
    'Flare linelist 1'
    >>> data["raster_fov"]
    '31.35x120'
    >>> data["raster_step"]
    np.float64(0.33)
    """

    def __init__(self, obsid) -> None:
        self.obsid = obsid
        data, options = self._read_obsid(obsid)
        super().__init__(data)
        self.options = options

    def __repr__(self) -> str:
        return (
            "IRIS OBS ID {obsid}\n"
            "----------------------\n"
            "Description:       {raster_fulldesc:>45}\n"
            "SJI filters:       {sjis:>45}\n"
            "SJI field of view: {sji_fov:>45}\n"
            "Exposure time:     {exptime:>45}\n"
            "Binning:           {binning:>45}\n"
            "FUV binning:       {fuv_binning:>45}\n"
            "SJI cadence:       {sji_cadence:>45}\n"
            "Compression:       {compression:>45}\n"
            "Linelist:          {linelist:>45}"
        ).format(**self)

    @staticmethod
    def _exptime_to_quant(exptime):
        """
        Converts an 'exptime' string (used in IRIS tables and OBS_DESC) to a
        Quantity instance in seconds.
        """
        if exptime == "Exposure 1s":
            return 1.0 * u.s
        return float(exptime.split(" x ")[1]) * u.s

    def _read_obsid(self, obsid):
        """
        Reads different fields from OBS ID number.
        """
        options = {}
        data = {"obsid": obsid}
        field_keys = {
            "Large linelist": "linelist",
            "Default compression": "compression",
            "Lossy compression": "compression",
            "Non-simultaneous readout": "readout",
            "SJI cadence default": "sji_cadence",
            "SJI cadence 10s": "sji_cadence",
            "FUV binned same as NUV": "fuv_binning",
            "Spatial x 1, Spectral x 1": "binning",
            "Exposure 1s": "exptime",
            "C II   Si IV   Mg II h/k   Mg II w   ": "sjis",
        }
        versions = [34, 36, 38, 40]
        if len(str(obsid)) != 10:
            msg = "Invalid OBS ID: must have 10 digits."
            raise ValueError(msg)
        # here choose between tables
        version = int(str(obsid)[:2])
        if version not in versions:
            msg = f"Invalid OBS ID: two first digits must one of {versions}"
            raise ValueError(msg)
        obsid = int(str(obsid)[2:])  # version digits are no longer needed
        with importlib.resources.as_file(importlib.resources.files("irispy") / f"data/v{version}-table10.csv") as path:
            table1 = pd.read_csv(path)
        with importlib.resources.as_file(
            importlib.resources.files("irispy") / f"data/v{version}-table2000.csv"
        ) as path:
            table2 = pd.read_csv(path)
        id_raster = int(str(obsid)[-2:])
        try:
            meta = table1.where(table1["OBS-ID"] == id_raster).dropna().iloc[0]
        except IndexError:
            msg = f"Invalid OBS ID: last two numbers must be between {table1['OBS-ID'].min()} and {table1['OBS-ID'].max()}"
            raise ValueError(
                msg,
            ) from None

        data["raster_step"] = meta["Raster step"]
        data["raster_fov"] = meta["Raster FOV"]
        data["spec_cadence"] = meta["Spectral cadence"]
        data["sji_fov"] = meta["SJI FOV"]
        data["raster_desc"] = meta["Description"]
        data["raster_fulldesc"] = f"{data['raster_desc']} {data['raster_fov']} {data['spec_cadence']}"
        field_ranges = np.concatenate(
            [  # find all dividers between fields
                table2.where(table2["OBS ID"] == 0).dropna(how="all").index,
                np.array([len(table2)]),
            ],
        )
        # field indices, start from largest and subtract
        for start, end in zip(field_ranges[-2::-1], field_ranges[:0:-1], strict=False):
            table = table2.iloc[start:end]
            for i in np.arange(start, end)[::-1]:
                index = i
                tmp = table["OBS ID"].loc[i]
                if (obsid - tmp) >= 0:
                    obsid -= tmp
                    break
            desc = table["Size + description"]
            # Save values for attributes but also table options as function of OBS ID
            if desc.iloc[0] in field_keys:
                attr_name = field_keys[desc.iloc[0]]
                if attr_name == "exptime":
                    opt = (self._exptime_to_quant(a) for a in list(desc.values))
                    opt = dict(zip(opt, table["OBS ID"], strict=False))
                    attr_value = self._exptime_to_quant(desc.loc[index])
                else:
                    opt = dict(zip(desc, table["OBS ID"], strict=False))
                    attr_value = desc.loc[index].strip()
                data[attr_name] = attr_value
                options[attr_name] = opt
        return data, options
